#ifdef KITSCHY_DEBUG_MEMORY
#include "debug_memorymanager.h"
#endif

#ifdef _WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "math.h"
#include "string.h"

#include "debug.h"

#include "GL/gl.h"
#include "SDL.h"
#include "SDL_mixer.h"

#include "BList.h"
#include "Symbol.h"
#include "2DCMC.h"
#include "auxiliar.h"
#include "GLTile.h"
#include "2DCMC.h"
#include "keyboardstate.h"
#include "randomc.h"
#include "VirtualController.h"
#include "SDL_glutaux.h"

#include "GLTManager.h"
#include "SoundManager.h"
#include "SFXManager.h"
#include "MusicManager.h"
#include "KObject.h"
#include "Knightmare.h"
#include "KnightmareCtnt.h"
#include "KnightmareApp.h"

#include "Binterface.h"


/* fps counter: */
extern int frames_per_sec;
extern int frames_per_sec_tmp;

extern int init_time;
extern bool show_fps;
extern bool fullscreen;

extern int difficulty;
extern int current_cycle;

KnightmareApp::KnightmareApp()
{
	m_font16=TTF_OpenFont("fonts/OLYMRG__.TTF",16);
	m_font32=TTF_OpenFont("fonts/OLYMRG__.TTF",32);

    m_state = STATE_MSX;
//    m_state = STATE_TITLEANIMATION;
//    m_state = STATE_TITLE;
//	m_state = STATE_GAMESTART;

	m_previous_state = m_state;
    m_state_cycle = 0;

    m_skip_intro_screens = false;

	m_bat_effect = 0;
	
	m_gamestart_state = 0;

    m_game_state = 0;
    m_n_players = 1;
    m_current_level = 1;
	m_n_lives_left_player1 = 3;
	m_n_lives_left_player2 = 3;

    m_gameover_state = 0;
    m_endsequence_state = 0;

	m_mouse_x=0;
	m_mouse_y=0;
	m_draw_mouse_status=0;
	m_draw_mouse=0.0f;
	m_mouse_movement_timmer=0;
	m_mouse_button=0;

    m_screen_dx = SCREEN_X;
    m_screen_dy = SCREEN_Y;
    fullscreen = false;

    m_GLTM = new GLTManager();

    m_SFXM = new SFXManager();
    m_SFXM->cache("sfx");
//	m_GLTM->cache();
    m_MusicM = new MusicManager();

    // default keyboard config
	m_joystick1_num = -1;
	m_joystick2_num = -1;
    m_keys_configuration1[KKEY_UP] = SDLK_UP;
    m_keys_configuration1[KKEY_RIGHT] = SDLK_RIGHT;
    m_keys_configuration1[KKEY_DOWN] = SDLK_DOWN;
    m_keys_configuration1[KKEY_LEFT] = SDLK_LEFT;
    m_keys_configuration1[KKEY_FIRE] = SDLK_SPACE;
    m_keys_configuration1[KKEY_QUIT] = SDLK_ESCAPE;

    m_keys_configuration2[KKEY_UP] = SDLK_w;
    m_keys_configuration2[KKEY_RIGHT] = SDLK_d;
    m_keys_configuration2[KKEY_DOWN] = SDLK_s;
    m_keys_configuration2[KKEY_LEFT] = SDLK_a;
    m_keys_configuration2[KKEY_FIRE] = SDLK_LCTRL;
    m_keys_configuration2[KKEY_QUIT] = SDLK_ESCAPE;

    m_sfx_volume = MIX_MAX_VOLUME;
    m_music_volume = 64;
    m_game = 0;

    // Create the virtual controllers for both players:
    m_vc_l.Add(new VirtualController(0));
    m_vc_l.Add(new VirtualController(1));
	m_num_joysticks = SDL_NumJoysticks();

	m_title_menu_ID=0;
	m_title_menu_name=0;
	m_title_menu_state = 0;
	m_title_entry_effect = 0.0f;
	m_title_menu_selected = 0;
	m_title_menu_action = -1;

	m_title_menu_redefining_key_player = -1;
	m_title_menu_redefining_key = -1;

    m_test_game = 0;
    load_configuration();
	load_hiscores();
	
	for(int i = 0;i<MAX_HIGHSCORES;i++) {
		m_highscore_scores_1player[i] = 0;
		m_highscore_names_1player[i] = 0;
		m_highscore_scores_2players[i] = 0;
		m_highscore_names_2players[i] = 0;
	}
}


KnightmareApp::~KnightmareApp()
{
	if (m_bat_effect!=0) delete m_bat_effect;

	TTF_CloseFont(m_font16);
	TTF_CloseFont(m_font32);

    if (m_game != 0) {
        delete m_game;
	    m_game = 0;
    } // if

    delete m_GLTM;
    delete m_SFXM;
    delete m_MusicM;

    if (m_title_menu_name!=0) delete m_title_menu_name;
    m_title_menu_name = 0;

    save_configuration();
	save_hiscores();
}


void KnightmareApp::MouseClick(int mousex,int mousey)
{
	m_mouse_click_x.Add(new int(mousex));
	m_mouse_click_y.Add(new int(mousey));
} /* KnightmareApp::MouseClick */


bool KnightmareApp::cycle(KEYBOARDSTATE *k)
{
    int old_state = m_state;

#ifdef __DEBUG_MESSAGES
    if (m_state_cycle == 0) {
        output_debug_message("First Cycle started for state %i...\n", m_state);
    }
#endif

    switch (m_state) {
        case STATE_SPLASH:
            m_state = splash_cycle(k);
            break;
        case STATE_MSX:
            m_state = msx_cycle(k);
            break;
        case STATE_KONAMI:
            m_state = konami_cycle(k);
            break;
        case STATE_TITLE:
            m_state = title_cycle(k);
            break;
        case STATE_TITLEANIMATION:
            m_state = titleanimation_cycle(k);
            break;
        case STATE_GAMESTART:
            m_state = gamestart_cycle(k);
            break;
        case STATE_GAME:
            m_state = game_cycle(k);
            break;
        case STATE_GAMEOVER:
            m_state = gameover_cycle(k);
            break;
        case STATE_ENDSEQUENCE:
            m_state = endsequence_cycle(k);
            break;
        case STATE_INTERLEVEL:
            m_state = interlevel_cycle(k);
            break;
        case STATE_HOWTOPLAY:
            m_state = howtoplay_cycle(k);
            break;
        case STATE_CREDITS:
            m_state = credits_cycle(k);
            break;
        case STATE_TEST:
            m_state = test_cycle(k);
            break;
        default:
            return false;
    }

    if (old_state == m_state) {
        m_state_cycle++;
    } else {
		if (m_state == STATE_GAMEOVER && m_game != 0) {
			// for now, we set a temporary player name. Ready for when, in the future, we have online hiscores:
			Symbol *tmp = new Symbol("playername");
			set_1phiscore(m_game->get_score(),tmp);
			delete tmp;
		}

        m_state_cycle = 0;

#ifdef __DEBUG_MESSAGES
        output_debug_message("State change: %i -> %i\n", old_state, m_state);
#endif
    }

    m_SFXM->next_cycle();
    m_MusicM->next_cycle();

	m_previous_state = old_state;

    return true;
}

void KnightmareApp::draw()
{
    float lightpos[4] = {0, -1000, -1000, 0};
    float tmpls[4] = {1.0F, 1.0F, 1.0F, 1.0};
    float tmpld[4] = {1.0F, 1.0F, 1.0F, 1.0};
    float tmpla[4] = {1.0F, 1.0F, 1.0F, 1.0};
    float ratio;

    m_screen_dx = SCREEN_X;
    m_screen_dy = SCREEN_Y;

    if (m_state_cycle == 0) {
#ifdef __DEBUG_MESSAGES
        output_debug_message("First Drawing cycle started for state %i...\n", m_state);
#endif
		// Never draw before the first cycle of a state has executed
		return;
    }

    /* Enable Lights, etc.: */
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // Really Nice Perspective Calculations
    glEnable(GL_LIGHT0);

    glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
    glLightfv(GL_LIGHT0, GL_AMBIENT, tmpla);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, tmpld);
    glLightfv(GL_LIGHT0, GL_SPECULAR, tmpls);

    glEnable(GL_LIGHTING);
    glEnable(GL_COLOR_MATERIAL);
    glShadeModel(GL_SMOOTH);
    glDisable(GL_CULL_FACE);
    glDisable(GL_SCISSOR_TEST);
    glDisable(GL_DEPTH_TEST);
    glPolygonMode(GL_FRONT, GL_FILL);
	glFrontFace(GL_CCW);

    glClearColor(0, 0, 0, 0.0);
    glViewport(0, 0, SCREEN_X, SCREEN_Y);
    ratio = (float)SCREEN_X / float(SCREEN_Y);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    my_gluPerspective(30.0, ratio, 1.0, 10240.0);
    my_gluLookAt(320, 240, PERSPECTIVE_DISTANCE, 320, 240, 0); /* for 640x480 better */

//    gluLookAt(320, 240, PERSPECTIVE_DISTANCE, 320, 240, 0, 0, -1, 0); /* for 640x480 better */

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    switch (m_state) {
        case STATE_SPLASH:
            splash_draw();
            break;
        case STATE_MSX:
            msx_draw();
            break;
        case STATE_KONAMI:
            konami_draw();
            break;
        case STATE_TITLE:
            title_draw();
            break;
        case STATE_TITLEANIMATION:
            titleanimation_draw();
            break;
        case STATE_GAMESTART:
            gamestart_draw();
            break;
        case STATE_GAME:
            game_draw();
            break;
        case STATE_GAMEOVER:
            gameover_draw();
            break;
        case STATE_ENDSEQUENCE:
            endsequence_draw();
            break;
        case STATE_INTERLEVEL:
            interlevel_draw();
            break;
        case STATE_HOWTOPLAY:
            howtoplay_draw();
            break;
        case STATE_CREDITS:
            credits_draw();
            break;
        case STATE_TEST:
            test_draw();
            break;
    }

    if (show_fps) {
        char tmp[80];
        sprintf(tmp, "video mem: %.4gmb - fps: %i", float(GLTile::get_memory_used()) / float(1024 * 1024), frames_per_sec);
		Binterface::print_center(tmp,m_font16,320, 460, 1,0,0,1);
    }

    glDisable(GL_BLEND);

    SDL_GL_SwapBuffers();
}


int KnightmareApp::screen_x(int x)
{
    return ((x * m_screen_dx) / SCREEN_X);
}


int KnightmareApp::screen_y(int y)
{
    return ((y*m_screen_dy) / SCREEN_Y);
}


void KnightmareApp::save_configuration(void)
{
    int i;
    FILE *fp;
    char cfg[255];
#ifdef _WIN32

    sprintf(cfg, "knightmare.cfg");
#else

    snprintf(cfg, 255, "%s/.knightmare.cfg", getenv("HOME"));
#endif //_WIN32

    fp = fopen(cfg, "w");
    if (fullscreen) {
        fprintf(fp, "fullscreen\n");
    } else {
        fprintf(fp, "windowed\n");
    }
    fprintf(fp, "%i %i\n" , m_sfx_volume, m_music_volume);
    fprintf(fp, "%i\n", difficulty);

    for (i = 0; i < 6; i++) {
        fprintf(fp, "%i ", m_keys_configuration1[i]);
    }

    for (i = 0; i < 6; i++) {
        fprintf(fp, "%i ", m_keys_configuration2[i]);
    }
    fprintf(fp, "\n");
    fclose(fp);
}


void KnightmareApp::load_configuration(void)
{
    FILE *fp;
    char cfg[255];
#ifdef _WIN32

    sprintf(cfg, "knightmare.cfg");
#else

    snprintf(cfg, 255, "%s/.knightmare.cfg", getenv("HOME"));
#endif //_WIN32

    fp = fopen(cfg, "r");
    if (fp == 0) {
        save_configuration();
    } else {
        char tmp_s[80];
        int i;

        // fullscreen / windowed
        fscanf(fp, "%s", tmp_s);
        if (strcmp(tmp_s, "fullscreen") == 0) {
            fullscreen = true;
        } else {
            fullscreen = false;
        }
        // volumes
        fscanf(fp, "%i %i", &m_sfx_volume, &m_music_volume);

        // difficulty
        fscanf(fp, "%i", &difficulty);

        // keyboard
        for (i = 0; i < 6; i++) {
            fscanf(fp, "%i", &(m_keys_configuration1[i]));
        }

        for (i = 0; i < 6; i++) {
            fscanf(fp, "%i", &(m_keys_configuration2[i]));
        }

        fclose(fp);
    }
}


void KnightmareApp::save_hiscores(void)
{
    FILE *fp;
    char cfg[255];
#ifdef _WIN32

    sprintf(cfg, "knightmare.hi");
#else

    snprintf(cfg, 255, "%s/.knightmare.hi", getenv("HOME"));
#endif //_WIN32

    fp = fopen(cfg, "w");

	// ...

    fclose(fp);
}

void KnightmareApp::load_hiscores(void)
{
    FILE *fp;
    char cfg[255];
#ifdef _WIN32

    sprintf(cfg, "knightmare.hi");
#else

    snprintf(cfg, 255, "%s/.knightmare.hi", getenv("HOME"));
#endif //_WIN32

    fp = fopen(cfg, "r");
    if (fp == 0) {
        save_hiscores();
    } else {

		// ...

        fclose(fp);
    }
}


void KnightmareApp::draw_square(int x1,int y1,int x2,int y2,float r,float g,float b,float a)
{
	glEnable(GL_COLOR_MATERIAL);

    glColor4f(r, g, b, a);
    glNormal3f(0.0, 0.0, 1.0);

    glBegin(GL_QUADS);
    glVertex3f(float(x1), float(y1), 0);
    glVertex3f(float(x1), float(y2), 0);
    glVertex3f(float(x2), float(y2), 0);
    glVertex3f(float(x2), float(y1), 0);
    glEnd();
} // draw_square


bool KnightmareApp::keyPress(int key,int player,KEYBOARDSTATE *k)
{
	if (player==0 || player==-1) {
		if (k->keyboard[m_keys_configuration1[key]] && !k->old_keyboard[m_keys_configuration1[key]]) return true;
		if (m_joystick1_num>=0) {
			if (k->keyboard[(k->joystick_0_pos + m_joystick1_num * k->joystick_size + 4)] &&
				!k->old_keyboard[(k->joystick_0_pos + m_joystick1_num * k->joystick_size + 4)]) return true;
		}
	}
	if (player==1 || player==-1) {
		if (k->keyboard[m_keys_configuration2[key]] && !k->old_keyboard[m_keys_configuration2[key]]) return true;
		if (m_joystick2_num>=0) {
			if (k->keyboard[(k->joystick_0_pos + m_joystick2_num * k->joystick_size + 4)] &&
				!k->old_keyboard[(k->joystick_0_pos + m_joystick2_num * k->joystick_size + 4)]) return true;
		}
	}

	return false;
}

bool KnightmareApp::keyPressed(int key,int player,KEYBOARDSTATE *k)
{
	if (player==0 || player==-1) {
		if (k->keyboard[m_keys_configuration1[key]]) return true;
		if (m_joystick1_num>=0) {
			if (k->keyboard[(k->joystick_0_pos + m_joystick1_num * k->joystick_size + 4)]) return true;
		}
	}
	if (player==1 || player==-1) {
		if (k->keyboard[m_keys_configuration2[key]]) return true;
		if (m_joystick2_num>=0) {
			if (k->keyboard[(k->joystick_0_pos + m_joystick2_num * k->joystick_size + 4)]) return true;
		}
	}

	return false;
}


int KnightmareApp::get_1phiscore(void)
{
	return m_highscore_scores_1player[0];
}


int KnightmareApp::get_2phiscore(void)
{
	return m_highscore_scores_2players[0];
}


void KnightmareApp::set_1phiscore(int newscore, Symbol *player_name)
{
	int position = -1;
	for(int i = 0;i<MAX_HIGHSCORES;i++) {
		if (newscore>m_highscore_scores_1player[i]) {
			
			if (m_highscore_names_1player[MAX_HIGHSCORES-1]!=0) delete m_highscore_names_1player[MAX_HIGHSCORES-1];
			for(int j = MAX_HIGHSCORES;j>i;j--) {
				m_highscore_names_1player[j] = m_highscore_names_1player[j-1];
				m_highscore_scores_1player[j] = m_highscore_scores_1player[j-1];
			}
			m_highscore_names_1player[i] = new Symbol(player_name);
			m_highscore_scores_1player[i] = newscore;
			return;
		}
	}
}


void KnightmareApp::set_2phiscore(int newscore, Symbol *player_name)
{
	int position = -1;
	for(int i = 0;i<MAX_HIGHSCORES;i++) {
		if (newscore>m_highscore_scores_2players[i]) {
			
			if (m_highscore_names_2players[MAX_HIGHSCORES-1]!=0) delete m_highscore_names_2players[MAX_HIGHSCORES-1];
			for(int j = MAX_HIGHSCORES;j>i;j--) {
				m_highscore_names_2players[j] = m_highscore_names_2players[j-1];
				m_highscore_scores_2players[j] = m_highscore_scores_2players[j-1];
			}
			m_highscore_names_2players[i] = new Symbol(player_name);
			m_highscore_scores_2players[i] = newscore;
			return;
		}
	}
}

